import { connect as apiConnect, Channel, Options, Connection } from 'amqplib/callback_api';
import retry from 'async-retry';


const RECONNECT_INTERVAL_MS = 300;

const PUBLISH_RETRY_COUNT = 10;

const PUBLISH_OPTIONS: Options.Publish = {
  contentType: 'application/json',
  persistent: true,
}

export interface OptionsType {
  projectId: string;
  url: string;
  exchangeId: string;
}

export const RabbitMQ = async (options: OptionsType) => {
  let channel: Channel | null = null;
  let connection: Connection | null = null;

  const generatePublishKey = (entityType: string, actionType: string, org: string) => {
    return `${options.projectId}.${org}.${entityType}.${actionType}`
  }

  const reconnect = () => {
    setTimeout(
      async () => {
        try {
          channel = await createChannel();
          console.log('RabbitMQ: connection was restored');
        } catch (e) {
          reconnect();
        }
      },
      RECONNECT_INTERVAL_MS
    )
  }
  
  const createChannel = async (): Promise<Channel> => {
    return new Promise((resolve, reject) => {
      apiConnect(options.url, (err, conn) => {
        if (!!err) {
          reject(new Error(`Failed to connect to RabbitMQ: ${err} (logging is disabled)`));
          return;
        }

        conn.on('error', (err) => {
          if (!channel) {
            return;
          }

          console.error(`RabbitMQ: connection error (${err}), attempting to reconnect... (logging is disabled)`);
          channel = null;
          reconnect();
        });
        
        conn.on('close', () => {
          if (!channel) {
            return;
          }

          console.error(`RabbitMQ: connection was closed, attempting to reconnect... (logging is disabled)`);
          channel = null;
          reconnect();
        });

        connection = conn;
        conn.createChannel((err, ch) => {
          if (!!err) {
            reject(new Error(`Failed to create channel: ${err} (logging is disabled)`));
            return;
          }

          ch.assertExchange(options.exchangeId, 'topic', { durable: true });

          channel = ch;
          resolve(ch);
        });
      });
    });
  }

  const close = async (): Promise<void> => {
    return new Promise((resolve, reject) => {
      if (!connection) {
        resolve();
        return;
      }

      connection.removeAllListeners();
      connection.close(err => {
        if (err) {
          reject(err);
        }

        resolve();
      });
    });
  };

  try {
    await createChannel();
    console.log(`RabbitMQ: connection was established to ${options.url!}`);
  } catch(e) {
    console.error(`RabbitMQ: failed to create channel: ${e}`);
  }

  return {
    publishEvent: async (entityType: string, actionType: string, data: object, org: string = 'common') => {
      const key = generatePublishKey(entityType, actionType, org);

      try {
        await retry(async (bail) => {
          if (!channel) {
            throw new Error('No channel');
          }

          if (!channel.publish(options.exchangeId, key, Buffer.from(JSON.stringify(data)), PUBLISH_OPTIONS)) {
            throw new Error('Write buffer full');
          }
        }, {
          retries: PUBLISH_RETRY_COUNT
        });
      } catch (e) {
        console.error(`RabbitMQ: failed to publish an event: ${e}`);
      }
    },
    close,
    isConnected: () => {
      return !!channel;
    }
  }
};
